#!/usr/bin/python3
# -*- coding: utf-8 -*-
################################################################################
#    Phoseg Copyright (C) 2012 Suizokukan
#    Contact: suizokukan at orange dot fr
#
#    This file is part of Phoseg.
#    Phoseg is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Phoseg is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with Phoseg.  If not, see <http://www.gnu.org/licenses/>.
################################################################################
"""
        ❏Phoseg❏ : phoseg/IPA/IPA_DATABASE.py

        * class IPA_DATABASE
"""

import sqlite3
import os
import re
import sys
import collections
import unicodedata
from phoseg.errors import errors
from phoseg.system import unicode_analysis

################################################################################
class IPA_DATABASE(object):
    """
        class IPA_DATABASE

        This class aims to store the details about IPA phonemes and to
        give some informations about it (e.g. the list of the IPA
        characters considered as vowels).

        An IPA_DATABASE object is normally initialized from a SQLite database
        (IPA_DATABASE().init_from_a_database); it's therefore
        possible to initialize the object from a textual representation
        of a SQLite database (IPA_DATABASE().init_from_a_text_database())
    """

    default__text_filename = os.path.join("phoseg","IPA","IPA_DATABASE.sql3.txt")
    default__database_filename = os.path.join("phoseg","IPA","IPA_DATABASE.sql3")

    # if you change this name please modify the documentation process
    # (e.g. build_the_Sphinx_translations.py)
    default__status_file_filename = os.path.join("phoseg","IPA","IPA_DATABASE.status.rst")

    str__entries_in_phonemes_table = "characters,identical_with,name,variety,sonority,diaphragm_air,glottis,glottis_voicing,glottis_air,epiglottis,velum_lowered,tongue_air,tongue_corona_lamina,tongue_corona_apex,tongue_corona_subapex,tongue_dorsum,tongue_radix,tongue_centrallateral,tongue_shape,vowel_height,vowel_backness,jaw_lowerteeth,jaw_lowerlip,rounded_lips"
    entries_in_phonemes_table = collections.namedtuple('PhonemeRecord', str__entries_in_phonemes_table)

    str__linked_entries_in_phonemes_table = "characters,identical_with,variety"
    linked_entries_in_phonemes_table = collections.namedtuple('PhonemeRecord_linked_entries', str__entries_in_phonemes_table)

    pattern__INSERT = re.compile("(-- ►\s+(?P<IPA_c>\S+)\s+\((?P<d1_before_equal>[0-9|A-Z|\s|+]*)\s*(=?\s*(?P<d1_after_equal>[0-9|A-Z|\s|+]*))?\)(\s+→\s+(?P<IPA_c2>\S+)\s+\((?P<d2_before_equal>[0-9|A-Z| |+]*)\s*=?(\s*(?P<d2_after_equal>[0-9|A-Z| |+]*)\))?)?\s*--\s*(--.*\n)*→ INSERT INTO phonemes \(\S*\) VALUES\s*\(\s+•characters•'(?P<IPA_c3>\S+)',\s+•identical_with•'(?P<identical_with>.*)',(\s+•name•'(?P<name>.*)')?)")

    #---------------------------------------------------------------------------
    def __del__(self):
        """
                IPA_DATABASE.__del__
        """
        # we have to be sure to close the connection :
        if self.connection is not None:
            self.connection.close()

    #---------------------------------------------------------------------------
    def __init__(self):
        """
                IPA_DATABASE.__init__
        """
        self.connection = None
        self.cursor = None

        # object defined by get_some_basic_informations() :
        self.list_of_IPA_characters = None

    #---------------------------------------------------------------------------
    def __repr__(self):
        """
                IPA_DATABASE.__repr__
        """
        return self.xstr(explore_the_database=True)

    #---------------------------------------------------------------------------
    def __str__(self):
        """
                IPA_DATABASE.__str__
        """
        return self.xstr(explore_the_database=False)

    #---------------------------------------------------------------------------
    def check_the_coherency_of_the_database(self):
        """
                IPA_DATABASE.check_the_coherency_of_the_database
                Raise an exception if something's wrong.
        """

        #. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .
        # test #1 : no pending "identical_with"
        # That test means that all link must exist in the database.
        #. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

        for source_characters,identical_with in self.cursor.execute('SELECT characters,identical_with FROM phonemes'):

            if identical_with != "":

                # we check if this name exists :
                query = self.cursor.execute('SELECT characters FROM phonemes WHERE name="{0}"'.format(identical_with))
                r = query.fetchone()
                if r is None:
                    raise errors.DataBaseError_incoherent_database("The phoneme '{0}' is identical with {1} which doesn't exist.".format(source_characters,identical_with))

        #. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .
        # test #2 : every name must be different
        #
        # NB : the name can be equal to None for the symbols designing
        #      another symbol (with the identical_with element).
        #. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .
        names = []

        query = self.cursor.execute('SELECT name FROM phonemes')
        for _name in query.fetchall():
            name = _name[0]

            if name is not None:
                if name!='None' and name not in names:
                    names.append(name)
                else:
                    raise errors.DataBaseError_incoherent_database("The name '{0}' is used at least two times".format(name))

        #. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .
        # test #3 : a link (identical_with=xxx) must have THREE and only THREE
        #           elements : (characters,identical_with,variety)
        #. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

        query = self.cursor.execute('SELECT {0} FROM phonemes WHERE identical_with!=""'.format(IPA_DATABASE.str__entries_in_phonemes_table))

        for emp in map(IPA_DATABASE.entries_in_phonemes_table._make, query.fetchall()):

            if emp.characters=="":
                raise errors.DataBaseError_incoherent_database("In a link (identical_with=xxx) 'characters' have to be set and cannot be empty; here characters={0}".format(emp.characters))

            if emp.variety=="":
                raise errors.DataBaseError_incoherent_database("In a link (identical_with=xxx) 'variety' have to be set and cannot be empty; here characters={0}; problem with variety={1}".format(emp.characters, emp.variety))

            # we check that every field except (characters,identical_with,variety) is None :
            for f in emp._fields:
                if f not in ('characters','identical_with','variety') and getattr(emp, f) is not None:
                    raise errors.DataBaseError_incoherent_database("In a link (identical_with=xxx) you can't store any information except (characters,identical_with,variety); characters='{0}'; problem with the string='{1}' in {2}.".format(emp.characters, getattr(emp, f),f))

    #---------------------------------------------------------------------------
    def check_the_consistency_of_a_text_database(self, filename):
        """
                IPA_DATABASE.check_the_consistency_of_a_text_database

                Raise an exception if the file doesn't fit the expected format.

                * check #1 : same number of "-- ► " and "→ INSERT".
                * check #2 : the header of each character (e.g. "-- ► ʪ (02AA)")
                  must be coherent.

                    * check #2a : no names' doubloon.
                    * check #2b : character in the comment line == character
                      in the INSERT statement.
                    * check #2c : coherency of the decomposition in the comment line

                * check #3 : the number of well-formed header must be equal to
                  the number of headers.

                * check #4 : we check the order of the insertion of the characters.

                * check #5 : every link (identical_with) must refer to an
                  existing entry
        """

        #. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .
        # check #1 : same number of "-- ► " and "→ INSERT".
        #. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .
        with open(filename, "r") as src:
            data = src.read()

            nbr_strokes = data.count("-- ►")
            nbr_insert = data.count("→ INSERT")
            if nbr_strokes != nbr_insert:
                raise errors.DataBaseError_incoherent_text_database("The number of strokes (-- ►) (={0}) differs from the number of INSERT statements (={1}).".format(nbr_strokes,nbr_insert))

            pseudo_headers = re.findall("(-- ►\s+\S+)",
                                        data)

        #. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .
        # check #2 : the header of each character (e.g. "-- ► ʪ (02AA)") must be
        # coherent.
        #. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

        # <decompositions> is used with check #4 :
        decompositions = []
        identical_with = []
        names = []
        with open(filename, "r") as src:
            data = src.read()

            nbr_headers = 0     #
            for header in re.finditer(IPA_DATABASE.pattern__INSERT, data):
                nbr_headers += 1

                # <header> = (e.g.) "-- ► ɪ̈ (026A 0308)"
                # we extract the IPA character(s) and its decomposition :
                groupdict = header.groupdict()
                IPA_c = groupdict["IPA_c"]
                IPA_c2 = groupdict["IPA_c2"]
                identical_with.append( groupdict["identical_with"] )
                new_name = groupdict["name"]

                # check #2a : no names' doubloon :
                if new_name is not None and new_name in names:
                    raise errors.DataBaseError_incoherent_text_database("Problem in the header : {0}; the name '{1}' has been defined previously.".format(header.group(),new_name))
                names.append( new_name )

                if nbr_headers-1 < len(pseudo_headers):
                    print(pseudo_headers[nbr_headers-1], "***", IPA_c, groupdict["name"], groupdict["d1_before_equal"])
                else:
                    print("????", IPA_c, groupdict["name"], "***", groupdict["d1_before_equal"])

                # check #2b :
                # character in the comment line == character in the INSERT statement.
                if groupdict['IPA_c'] != groupdict['IPA_c3']:
                    raise errors.DataBaseError_incoherent_text_database("Problem in the header : {0}; character in the comment ({1}) is different from the character in the INSERT statement ({2}).".format(header.group(),groupdict['IPA_c'],groupdict['IPA_c3']))

                # check #2c : coherency of the decomposition in the comment line :
                d = groupdict["d1_before_equal"].replace('+','')
                d = d.replace('=','')
                d = d.replace(' ','')
                c = unicode_analysis.return_an_analysis_of_a_string(IPA_c,raw_number_only=True)
                c = c.replace(' ','')
                if d != c:
                    raise errors.DataBaseError_incoherent_text_database("Problem in the header (1) : {0}; given decomposition={1} but IPA character is analysed as {2}.".format(header.group(),d,c))

                if groupdict["d2_before_equal"] is not None:
                    d = groupdict["d2_before_equal"].replace('+','')
                    d = d.replace('=','')
                    d = d.replace(' ','')
                    c = unicode_analysis.return_an_analysis_of_a_string(IPA_c2,raw_number_only=True)
                    c = c.replace(' ','')
                    if d != c:
                        raise errors.DataBaseError_incoherent_text_database("Problem in the header (2) : {0}; given decomposition={1} but IPA character is analysed as {2}.".format(header.group(),d,c))

                # we add the current character to <decompositions> :
                decompositions.append(groupdict["d1_before_equal"])
                decompositions[-1] = decompositions[-1].replace('+', ' ')
                decompositions[-1] = decompositions[-1].split(" ")

            #. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .
            # check #3 : the number of well-formed header must be equal to the
            # number of headers.
            #. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .
            if nbr_headers!=nbr_strokes:
                raise errors.DataBaseError_incoherent_text_database("The number of strokes (►) (={0}) differs from the number of well-formed headers (={1}).".format(nbr_strokes,nbr_headers))

            #. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .
            # check #4 : we check the order of the insertion of the characters;
            # the decomposition's numbers must appear in increasing order :
            # OK : 0078, then 0079 003C 0048, then 0079 003C, then 0079, then 00AA.
            # NOT :0078, then 0078 00AC
            #. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .
            for i,decomposition in enumerate(decompositions):
                if i>1:
                    # e.g. :    decompositions[i] = [ 1234, 00AB, 0099 ]
                    #        -> sub_d = [ [1234], [1234,00AB], ]
                    sub_d = [ decomposition[:index] for index in [r for r in range(1,len(decomposition))] ]

                    # we search if elements of <sub_d> appear in <decompositions>
                    # before <i>.

                    for i2 in range(0,i-1):
                        if decompositions[i2] in sub_d:
                            raise errors.DataBaseError_incoherent_text_database("Misplaced characters : {0}(rank #{1}) is before {2}(rank {3}).".format(decompositions[i2],i2,sub_d,i))

            #. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .
            # check #5 : every link (identical_with) must refer to an existing
            # entry
            #. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .
            for link in identical_with:
                if link != "" and link not in names:
                    raise errors.DataBaseError_incoherent_text_database("Pending link : {0}".format(link))

    #---------------------------------------------------------------------------
    def get_some_basic_informations(self):
        """
                IPA_DATABASE.get_some_basic_informations

                Initialize some usefull variables.
        """
        # the objects in this object must be ordered ! No dictionary in it !
        self.list_of_IPA_characters = { "vowels":[],
                                        "vowels_no_links":[],
                                        "consonants":[],
                                        "consonants_no_links":[] }

        for characters in self.cursor.execute('SELECT characters FROM phonemes WHERE variety="vowel"'):
            self.list_of_IPA_characters["vowels"].append( characters[0] )

        for characters in self.cursor.execute('SELECT characters FROM phonemes WHERE variety="vowel" AND identical_with==""'):
            self.list_of_IPA_characters["vowels_no_links"].append( characters[0] )

        for characters in self.cursor.execute('SELECT characters FROM phonemes WHERE variety="consonant"'):
            self.list_of_IPA_characters["consonants"].append( characters[0] )

        for characters in self.cursor.execute('SELECT characters FROM phonemes WHERE variety="consonant" AND identical_with==""'):
            self.list_of_IPA_characters["consonants_no_links"].append( characters[0] )

    #---------------------------------------------------------------------------
    def get_the_informations_about_the_database(self):
        """
                IPA_DATABASE.get_the_informations_about_the_database

                Return (tables, phonemes_infos, phonemes_rows, links )
        """

        # informations about the table(s) :
        tables = []
        for n in self.cursor.execute('SELECT name FROM sqlite_master where TYPE="table" ORDER BY NAME'):
            tables.append(n[0])

        self.cursor.execute( "PRAGMA table_info(vowels)" )

        # informations about the table 'phonemes' :
        phonemes_rows = []
        for n in self.cursor.execute( "PRAGMA table_info(phonemes)" ):
            phonemes_rows.append(n[1])

        # phonemes_infos looks like something like :
        # phonemes_infos = {"characters"                 : set(),
        #                   "identical_with"             : set(),
        #                   "name"                       : set(),
        #                   .....................................
        phonemes_infos = dict( (field,set()) for field in IPA_DATABASE.entries_in_phonemes_table._fields )

        query = self.cursor.execute('SELECT {0} FROM phonemes'.format(IPA_DATABASE.str__entries_in_phonemes_table))
        for emp in map(IPA_DATABASE.entries_in_phonemes_table._make, query.fetchall()):

            for field in IPA_DATABASE.entries_in_phonemes_table._fields:

                # phonemes_infos['characters'].add(emp.characters)
                # phonemes_infos['identical_with'].add(emp.identical_with)
                # phonemes_infos['name'].add(emp.name)
                # .............................................
                phonemes_infos[field].add( getattr(emp, field ))

        # links stored in the database (c1 → c2) :
        links = []
        query = self.cursor.execute('SELECT {0} FROM phonemes'.format(IPA_DATABASE.str__linked_entries_in_phonemes_table))
        for characters,identical_with,variety in query:
            if identical_with != "":
                links.append( (characters, identical_with,variety) )

        return (tables,
                phonemes_infos,
                phonemes_rows,
                links )

    #---------------------------------------------------------------------------
    def informations_about_a_phoneme(self, source_characters):
        """
                IPA_DATABASE.informations_about_a_phoneme

                This function respects the type stored in the database.
                E.g. if an integer has been stored the function returns
                an integer.

                This function returns a dictionary.
        """

        # is this phoneme identical with another phoneme ?
        query0 = self.cursor.execute('SELECT identical_with FROM phonemes where characters="{0}"'.format(source_characters))
        # <identical_with> : ('strings',) if there's an "identical_with",
        # or ('',) if no "identical_with".
        identical_with = query0.fetchone()[0]
        if identical_with!='':
            # there's 'identical_with' value, so we have to deal with it and follow the link.
            # <identical_with> contains the name of the real phoneme.
            query1 = self.cursor.execute('SELECT characters FROM phonemes WHERE name="{0}"'.format(identical_with))
            r = query1.fetchone()
            if r is None:
                raise errors.DataBaseError_incoherent_database("The phoneme '{0}' is identical with {1} which doesn't exist.".format(source_characters,identical_with))
            c = r[0]

        else:
            # no <identical_with> value so we can go on with the current phoneme :
            c = source_characters

        query2 = self.cursor.execute('SELECT {0} FROM phonemes where characters="{1}"'.format(IPA_DATABASE.str__entries_in_phonemes_table,
                                                                                              c))
        for emp in map( IPA_DATABASE.entries_in_phonemes_table._make, query2.fetchall() ):

            # we create the resulting dictionary, looking something like :
            # res = { "characters" : emp.characters, ... }
            res = dict( (field,getattr(emp,field)) for field in emp._fields )

        return res

    #---------------------------------------------------------------------------
    def init_from_a_database(self,
                             filename=default__database_filename):
        """
                IPA_DATABASE.init_from_a_database
        """

        if os.path.exists(filename):
            self.open_a_connection(filename)
            self.get_some_basic_informations()

        else:
            raise errors.DataBaseError_nodatabase(filename)

        return self

    #---------------------------------------------------------------------------
    def init_from_a_text_database(self,
                                  text_filename=default__text_filename,
                                  database_filename=default__database_filename):
        """
                IPA_DATABASE.init_from_a_text_database

                Init the database by reading a text file and by testing the
                coherency of the result.
        """

        # we check the consistency of the text file :
        # An exception is raised if the file doesn't fit the expected format.
        self.check_the_consistency_of_a_text_database(text_filename)

        # we read the commands (they all begin with an arrow, '→') :
        commands = []
        with open(text_filename, 'r') as textfile:

            for line in textfile:
                # we get rid of the comments (they all begin with '--') :
                if line[0:2]!="--":
                    if line[0]=='→':
                        commands.append( line[2:] )
                    else:
                        if len(commands)>0:
                            commands[-1] += line

        # we get rid of the points, like : (•diaphragm•) :
        commands = [ re.sub('•\w*•','',c) for c in commands]

        # we create the database by executing the commands :
        if os.path.exists(database_filename):
            os.remove(database_filename)

        self.open_a_connection(database_filename)

        try:
            for command in commands:
                self.cursor.execute(command)
                self.connection.commit()

        except sqlite3.OperationalError as e:
            sys.stderr.write("-"*80+"\n")
            sys.stderr.write("An exception occured while executing the following command :"+"\n")
            sys.stderr.write("-"*80+"\n")
            sys.stderr.write(command+"\n")
            sys.stderr.write("-"*80+"\n")
            raise e

        except sqlite3.IntegrityError as e:
            sys.stderr.write("-"*80+"\n")
            sys.stderr.write("An exception occured while executing the following command :"+"\n")
            sys.stderr.write("-"*80+"\n")
            sys.stderr.write(command+"\n")
            sys.stderr.write("-"*80+"\n")
            raise e

        self.get_some_basic_informations()

        self.check_the_coherency_of_the_database()

        self.refresh_the_status_file()

        return self

    #---------------------------------------------------------------------------
    def get_a_character_by_seeking_its_name(self, name):
        """
                IPA_DATABASE.get_a_character_by_seeking_its_name

                Return None if <name> doesn't exist in the database or
                a string.
        """

        query = self.cursor.execute('SELECT characters FROM phonemes where name="{0}"'.format(name))
        res = query.fetchone()

        # res : None / ('c',)
        if res is not None:
            res = res[0]

        return res

    #---------------------------------------------------------------------------
    def open_a_connection(self, filename):
        """
                IPA_DATABASE.open_a_connection
        """
        self.connection = sqlite3.connect(filename)
        self.cursor = self.connection.cursor()

    #---------------------------------------------------------------------------
    def refresh_the_status_file(self):
        """
                IPA_DATABASE.refresh_the_status_file

                Overwrite the 'status' file written with the RST format.

                If you change the strings appearing in the file beware of
                the translations ! The status file is a documentation file.
        """
        with open(IPA_DATABASE.default__status_file_filename, "w") as dest:

            tables, phonemes_infos, phonemes_rows, links = self.get_the_informations_about_the_database()

            dest.write(".. |br| raw:: html\n")
            dest.write("\n")
            dest.write("   <br />\n")
            dest.write("\n")

            dest.write("==============================="+"\n")
            dest.write("status file of the IPA database"+"\n")
            dest.write("==============================="+"\n")

            dest.write("\n")
            dest.write("------"+"\n")
            dest.write("tables"+"\n")
            dest.write("------"+"\n")

            for t in tables:
                dest.write("* {0} |br| \n".format(t))

            dest.write("\n")
            dest.write("----------------------"+"\n")
            dest.write("phonemes' informations"+"\n")
            dest.write("----------------------"+"\n")

            phonemes_infos__str = ""
            for field,data in phonemes_infos.items():
                phonemes_infos__str += " |br| \n\n* {0}:{1}".format(field,data)

            dest.write("{0} |br| \n".format(phonemes_infos__str))

            dest.write("\n")
            dest.write("-----"+"\n")
            dest.write("links"+"\n")
            dest.write("-----"+"\n")

            for characters,identical_with,variety in links:
                dest.write("* {0} → {2} |br|  {1}\n".format(characters,
                                                            unicode_analysis.return_an_analysis_of_a_string(characters),
                                                            identical_with))

            dest.write("\n")
            dest.write("--------------------------"+"\n")
            dest.write("consonants (without links)"+"\n")
            dest.write("--------------------------"+"\n")
            consonants = []
            for characters,name in self.cursor.execute('SELECT characters,name FROM phonemes WHERE variety="consonant" AND identical_with=""'):
                consonants.append( (name, characters) )
            consonants.sort()
            for name,c in consonants:
                dest.write("* {0} : {1} |br|  {2}\n".format(name,
                                                            c,
                                                            unicode_analysis.return_an_analysis_of_a_string(c)))

            dest.write("\n")
            dest.write("--------------------------"+"\n")
            dest.write("vowels (without links)"+"\n")
            dest.write("--------------------------"+"\n")
            vowels = []
            for characters,name in self.cursor.execute('SELECT characters,name FROM phonemes WHERE variety="vowel" AND identical_with=""'):
                vowels.append( (name, characters) )
            vowels.sort()
            for name,c in vowels:
                dest.write("* {0} : {1} |br|  {2}\n".format(name,
                                                            c,
                                                            unicode_analysis.return_an_analysis_of_a_string(c)))

    #---------------------------------------------------------------------------
    def sqlorder(self, order):
        """
                IPA_DATABASE.sqlorder

                To be used only as an interface for the user.
        """
        return self.cursor.execute(order)

    #---------------------------------------------------------------------------
    def xstr(self,explore_the_database):
        """
                IPA_DATABASE.xstr

                explore_the_database : (bool)
        """

        vowels = "{0} vowel(s) =".format(len(self.list_of_IPA_characters["vowels"]))+\
                 "/".join( self.list_of_IPA_characters["vowels"] )
        consonants = "{0} consonant(s) =".format(len(self.list_of_IPA_characters["consonants"]))+\
                     "/".join( self.list_of_IPA_characters["consonants"] )

        if not explore_the_database:
            return "\n".join( (vowels, consonants) )

        tables, phonemes_infos, phonemes_rows, links = self.get_the_informations_about_the_database()

        phonemes_infos__str = ""
        for p in phonemes_infos.items():
            phonemes_infos__str += "\n  + {0}:{1}".format(p[0],str(p[1]))

        res = "* tables : {0}".format(tables)+"\n"+\
              "\n".join( (vowels, consonants) )+"\n"+\
              "* phonemes_rows={0}".format(phonemes_rows)+"\n"+\
              "* phonemes_infos={0}".format( phonemes_infos__str )+"\n"+\
              "* links = {0}".format( links )

        return res




if __name__ == "__main__":

    import argparse

    parser = argparse.ArgumentParser(description="IPA_DATABASE.py is a part of the 'Phoseg' project. Use this file to create the IPA database.",
                                     epilog="Phoseg's author : Suizokukan (suizokukan@orange.fr)",
                                     formatter_class=argparse.ArgumentDefaultsHelpFormatter
                                    )

    parser.add_argument('--text2db',
                        action="store_true",
                        help="create the database from a text file",
                        default = False)

    parser.add_argument('--showdb',
                        action="store_true",
                        help="Show informations about the database that has been created",
                        default = False)

    parser.add_argument('--srcfile',
                        type=str,
                        help="name of the source file",
                        default = "-")

    args = parser.parse_args()

    # let's create a new database from a text file :
    if args.text2db:
        if args.srcfile == "-":
            db = IPA_DATABASE().init_from_a_text_database()
        else:
            db = IPA_DATABASE().init_from_a_text_database( args.srcfile )

        if args.showdb:
            print(repr(db))
