#!/usr/bin/python3
# -*- coding: utf-8 -*-
################################################################################
#    Phoseg Copyright (C) 2012 Suizokukan
#    Contact: suizokukan at orange dot fr
#
#    This file is part of Phoseg.
#    Phoseg is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Phoseg is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with Phoseg.  If not, see <http://www.gnu.org/licenses/>.
################################################################################
"""
    ❏Phoseg❏ : phoseg/phoneticsegment/phoneticsegment.py

    With a <PhoneticSegment> object we can store <Syllable> objects,
    made up of <Phoneme> objects madeup of one or more <SimplePhoneme>
    objects.
"""

import unicodedata
import re

# IPA stuff :
from phoseg.IPA.IPA_data import IPA_data as IPA_data

# error classes :
from phoseg.errors import errors as errors

################################################################################
class Syllable(object):
    """
        class Syllable : a syllable is defined by :

        * its tone
        * its stress
        * its onset (Phoneme object)
        * its peak (Phoneme object)
        * its coda (Phoneme object)
    """

    #---------------------------------------------------------------------------
    def __init__(self, tone, stress, onset, peak, coda):
        """
                Syllable.__init__
        """
        self.tone = tone        # (int)
        self.stress = stress    # (int)
        # three lists of Phoneme objects :
        self.onset = onset
        self.peak = peak
        self.coda = coda

    #---------------------------------------------------------------------------
    def __iter__(self):
        """
                Syllable.__iter__

                Return the Phoneme objects in self>, one by one.
        """
        for p in self.onset:
            yield p

        for p in self.peak:
            yield p

        for p in self.coda:
            yield p


    #---------------------------------------------------------------------------
    def __repr__(self):
        """
                Syllable.__repr__
        """
        return "tone={0}; stress={1}; ([{2}],[{3}],[{4}])".format(self.tone,
                                                                  self.stress,
                                                                  "/".join( [repr(e) for e in self.onset] ),
                                                                  "/".join( [repr(e) for e in self.peak] ),
                                                                  "/".join( [repr(e) for e in self.coda] ))

    #---------------------------------------------------------------------------
    def __str__(self):
        """
                Syllable.__str__
        """
        return "tone={0}; stress={1}; ([{2}],[{3}],[{4}])".format(self.tone,
                                                                  self.stress,
                                                                  "/".join( [str(e) for e in self.onset] ),
                                                                  "/".join( [str(e) for e in self.peak] ),
                                                                  "/".join( [str(e) for e in self.coda] ))

    #---------------------------------------------------------------------------
    def checked_syllable(self):
        """
                Syllable.checked_syllable
        """
        return not self.open_syllable()

    #---------------------------------------------------------------------------
    def closed_syllable(self):
        """
                Syllable.closed_syllable
        """
        return not self.open_syllable()

    #---------------------------------------------------------------------------
    def free_syllable(self):
        """
                Syllable.free_syllable
        """
        return self.open_syllable()

    #---------------------------------------------------------------------------
    def get_structure(self):
        """
                Syllable.get_structure

                Return a string ('V', 'CCVCCC', ...)
        """
        res = ""

        for p in iter(self):
            if p.is_vocalic():
                res += "V"
            else:
                res += "C"
        return res

    #---------------------------------------------------------------------------
    def open_syllable(self):
        """
                Syllable.open_syllable
        """
        return len(self.coda)==0

################################################################################
class Phoneme(list):
    """
        class Phoneme

        A <Phoneme> object is made of one or more <SimplePhoneme> objects.
    """

    #---------------------------------------------------------------------------
    def __init__(self, *Phoneme_objects):
        """
                Phoneme.__init__
        """
        for p in Phoneme_objects:
            self.append(p)

    #---------------------------------------------------------------------------
    def __repr__(self):
        """
                Phoneme.__repr__
        """
        return " & ".join( repr(simplephoneme) for simplephoneme in self )

    #---------------------------------------------------------------------------
    def __str__(self):
        """
                Phoneme.__str__
        """
        return " & ".join( str(simplephoneme) for simplephoneme in self )

    #---------------------------------------------------------------------------
    def is_vocalic(self):
        """
                Phoneme.is_vocalic

                Return a boolean : True if all SimplePhoneme objects are
                vocalic.
        """
        res = True

        for simplephoneme in self:
            if not simplephoneme.is_vocalic():
                res = False
                break

        return res

################################################################################
class SimplePhoneme(object):
    """
        class SimplePhoneme

        A phoneme is much much complex in reality and this class is based on
        the IPA characteristics of a phoneme.
        Two level of description : IPA description and "abstract" description.

        the __init__ function HAS NO ARGUMENT : a SimplePhoneme object must be
        described with the init_from_* functions.
    """


    #---------------------------------------------------------------------------
    def __repr__(self):
        """
                SimplePhoneme.__repr__
        """
        return self.xstr(IPA_only=False)

    #---------------------------------------------------------------------------
    def __str__(self):
        """
                SimplePhoneme.__str__
        """
        return self.xstr(IPA_only=True)

    #---------------------------------------------------------------------------
    def init_from_the_IPA_characteristics(self,characteristics):
        """
                SimplePhoneme.init_from_the_IPA_characteristics

                characteristics : { "articulation": "p",
                                    "aspirated":"ʰ" / "" }

                All the characteristics are strings; a string can be empty.
        """

        def find_a_characteristic(src):
            """
                Return a string; if <src> isn't in <characteristics>, return
                an empty string.
            """
            if src in characteristics:
                return characteristics[src]
            else:
                return ""

        def find_a_numeric_characteristic(src):
            """
                Return an integer greater or equal to 0
            """

            if src in characteristics:
                return len(characteristics[src])
            else:
                return 0

        # //////////////////////////////////
        # / IPA description of the phoneme /
        # //////////////////////////////////
        # one character : iyɨ (...) ɱmn (...) without any diacritic
        self.IPA_articulation           = find_a_characteristic('articulation')
        # diacritics written under a character :
        self.IPA_advanced               = find_a_numeric_characteristic('advanced')
        self.IPA_advanced_tongue_root   = find_a_numeric_characteristic('advanced_tongue_root')
        self.IPA_apical                 = find_a_numeric_characteristic('apical')
        self.IPA_breathy_voice          = find_a_numeric_characteristic('breathy_voice')
        self.IPA_creaky_voice           = find_a_numeric_characteristic('creaky_voice')
        self.IPA_dental                 = find_a_numeric_characteristic('dental')
        self.IPA_laminal                = find_a_numeric_characteristic('laminal')
        self.IPA_less_rounded           = find_a_numeric_characteristic('less_rounded')
        self.IPA_linguolabial_consonant = find_a_numeric_characteristic('linguolabial_consonant')
        self.IPA_lowered                = find_a_numeric_characteristic('lowered')
        self.IPA_mid_centralized        = find_a_numeric_characteristic('mid_centralized')
        self.IPA_more_rounded           = find_a_numeric_characteristic('more_rounded')
        self.IPA_raised                 = find_a_numeric_characteristic('raised')
        self.IPA_retracted              = find_a_numeric_characteristic('retracted')
        self.IPA_retracted_tongue_root  = find_a_numeric_characteristic('retracted_tongue_root')
        self.IPA_syllabic               = find_a_numeric_characteristic('syllabic')
        self.IPA_voiced                 = find_a_numeric_characteristic('voiced')
        self.IPA_voiceless              = find_a_numeric_characteristic('voiceless')
        # diacritics written over a character :
        self.IPA_centralized_vowel      = find_a_numeric_characteristic('centralized_vowel')
        self.IPA_nasalized              = find_a_numeric_characteristic('nasalized')
        self.IPA_no_audible_release     = find_a_numeric_characteristic('no_audible_release')
        # diacritics written beside a character :
        self.IPA_aspirated              = find_a_numeric_characteristic('aspirated')
        self.IPA_labialized             = find_a_numeric_characteristic('labialized')
        self.IPA_labio_palatalized      = find_a_numeric_characteristic('labio_palatalized')
        self.IPA_lateral_release        = find_a_numeric_characteristic('lateral_release')
        self.IPA_length                 = find_a_characteristic('length')
        self.IPA_nasal_release          = find_a_numeric_characteristic('nasal_release')
        self.IPA_palatalized            = find_a_numeric_characteristic('palatalized')
        self.IPA_pharyngealized         = find_a_numeric_characteristic('pharyngealized')
        self.IPA_rhotacized             = find_a_numeric_characteristic('rhotacized')
        self.IPA_velarized              = find_a_numeric_characteristic('velarized')
        self.IPA_voiced_aspirated       = find_a_numeric_characteristic('voiced_aspirated')
        self.IPA_glottalized            = find_a_numeric_characteristic('glottalized')
        self.IPA_extra_short            = find_a_numeric_characteristic('extra_short')
        # special diacritics :
        self.IPA_non_syllabic           = find_a_numeric_characteristic('non_syllabic')
        self.IPA_affricate              = find_a_numeric_characteristic('affricate')

        # We take in account the equivalences :
        self.modify_the_IPA_characteristics_along_equivalences_between_IPA_strings()

        # ///////////////////////////////////////
        # / abstract description of the phoneme /
        # ///////////////////////////////////////
        self.init_the_abstract_description()

        return self

    #---------------------------------------------------------------------------
    def is_vocalic(self):
        """
                SimplePhoneme.is_vocalic

                Return a boolean
        """
        return self.abstract_description['class'] == 'vowel'

    #---------------------------------------------------------------------------
    def modify_the_IPA_characteristics_along_equivalences_between_IPA_strings(self):
        """
                SimplePhoneme.modify_the_IPA_characteristics_along_equivalences_between_IPA_strings

                Modify IN PLACE the IPA characteristics of the object along some
                equivalences in the IPA strings. E.g. a+centralized_vowel = ä


                PLEASE MODIFY THE DOCUMENTATION IF YOU MODIFY THIS FUNCTION :
                documentation/Sphinx/manually_generated_RST_files/IPA_characters
        """

        # a+centralized_vowel → ä
        if self.IPA_articulation=="a" and self.IPA_centralized_vowel==1:
            self.IPA_articulation      = "ä"
            self.IPA_centralized_vowel = 0

        # o+centralized_vowel → ö
        if self.IPA_articulation=="o" and self.IPA_centralized_vowel==1:
            self.IPA_articulation      = "ö"
            self.IPA_centralized_vowel = 0

        # ɔ+centralized_vowel → ɞ
        if self.IPA_articulation=="ɔ" and self.IPA_centralized_vowel==1:
            self.IPA_articulation      = "ɞ"
            self.IPA_centralized_vowel = 0

        # e+centralized_vowel → ɘ
        if self.IPA_articulation=="e" and self.IPA_centralized_vowel==1:
            self.IPA_articulation      = "ɘ"
            self.IPA_centralized_vowel = 0

        # e+lowered → e̞
        if self.IPA_articulation=="e" and self.IPA_lowered==1:
            self.IPA_articulation = "e̞"
            self.IPA_lowered = 0

        # ø+lowered → ø̞
        if self.IPA_articulation=="ø" and self.IPA_lowered==1:
            self.IPA_articulation = "ø̞"
            self.IPA_lowered = 0

        # ɤ+lowered → ɤ̞
        if self.IPA_articulation=="ɤ" and self.IPA_lowered==1:
            self.IPA_articulation = "ɤ̞"
            self.IPA_lowered = 0

        # o+lowered → o̞
        if self.IPA_articulation=="o" and self.IPA_lowered==1:
            self.IPA_articulation = "o̞"
            self.IPA_lowered = 0

        # m̪ → ɱ
        if self.IPA_articulation=="m" and self.IPA_dental==1:
            self.IPA_articulation = "ɱ"
            self.IPA_dental = 0

    #---------------------------------------------------------------------------
    def init_the_abstract_description(self):
        """
                SimplePhoneme.init_the_abstract_description
        """

        self.abstract_description = {}

        # articulation → class of phoneme

        if self.IPA_articulation in IPA_data.IPA_vowels:
            # we have a vowel :
            self.abstract_description['class'] = 'vowel'

        elif self.IPA_articulation in IPA_data.IPA_consonants:
            # we have a vowel :
            self.abstract_description['class'] = 'consonant'

        # length of the phoneme ?
        self.abstract_description['length'] = IPA_data.IPA_length__character_to_value[self.IPA_length]

        # voiced character ?
        if self.IPA_voiced:
            self.abstract_description['voiceless'] = False
        elif self.IPA_voiceless:
            self.abstract_description['voiceless'] = True
        else:
            self.abstract_description['voiceless'] = IPA_data.database.informations_about_a_phoneme(self.IPA_articulation)["glottis_voicing"]

    #---------------------------------------------------------------------------
    def xstr(self, IPA_only):
        """
                SimplePhoneme.xstr

                IPA_only : (bool)

                Return a string describing the object. This function is called
                by the __repr__ and __str__ functions.
        """

        def format_a_numeric_characteristic(num,description):
            if num==0:
                return ""
            elif num==1:
                return description
            else:
                return description+("*"*num)

        res = "{0}".format(self.IPA_articulation)

        res += format_a_numeric_characteristic(self.IPA_advanced, "+advanced")
        res += format_a_numeric_characteristic(self.IPA_advanced_tongue_root, "+advanced_tongue_root")
        res += format_a_numeric_characteristic(self.IPA_apical, "+apical")
        res += format_a_numeric_characteristic(self.IPA_breathy_voice, "+breathy_voice")
        res += format_a_numeric_characteristic(self.IPA_creaky_voice, "+creaky_voice")
        res += format_a_numeric_characteristic(self.IPA_dental, "+dental")
        res += format_a_numeric_characteristic(self.IPA_laminal, "+laminal")
        res += format_a_numeric_characteristic(self.IPA_less_rounded, "+less_rounded")
        res += format_a_numeric_characteristic(self.IPA_linguolabial_consonant, "+linguolabial_consonant")
        res += format_a_numeric_characteristic(self.IPA_lowered, "+lowered")
        res += format_a_numeric_characteristic(self.IPA_mid_centralized, "+mid_centralized")
        res += format_a_numeric_characteristic(self.IPA_more_rounded, "+more_rounded")
        res += format_a_numeric_characteristic(self.IPA_raised, "+raised")
        res += format_a_numeric_characteristic(self.IPA_retracted, "+retracted")
        res += format_a_numeric_characteristic(self.IPA_retracted_tongue_root, "+retracted_tongue_root")
        res += format_a_numeric_characteristic(self.IPA_syllabic, "+syllabic")
        res += format_a_numeric_characteristic(self.IPA_voiced, "+voiced")
        res += format_a_numeric_characteristic(self.IPA_voiceless, "+voiceless")
        res += format_a_numeric_characteristic(self.IPA_centralized_vowel, "+centralized_vowel")
        res += format_a_numeric_characteristic(self.IPA_nasalized, "+nasalized")
        res += format_a_numeric_characteristic(self.IPA_no_audible_release, "+no_audible_release")
        res += format_a_numeric_characteristic(self.IPA_aspirated, "+aspirated")
        res += format_a_numeric_characteristic(self.IPA_labialized, "+labialized")
        res += format_a_numeric_characteristic(self.IPA_labio_palatalized, "+labio_palatalized")
        res += format_a_numeric_characteristic(self.IPA_lateral_release, "+lateral_release")
        if self.IPA_length!="":
            res += self.IPA_length
        res += format_a_numeric_characteristic(self.IPA_nasal_release, "+nasal_release")
        res += format_a_numeric_characteristic(self.IPA_palatalized, "+palatalized")
        res += format_a_numeric_characteristic(self.IPA_pharyngealized, "+pharyngealized")
        res += format_a_numeric_characteristic(self.IPA_rhotacized, "+rhotacized")
        res += format_a_numeric_characteristic(self.IPA_velarized, "+velarized")
        res += format_a_numeric_characteristic(self.IPA_voiced_aspirated, "+voiced_aspirated")
        res += format_a_numeric_characteristic(self.IPA_glottalized, "+glottalized")
        res += format_a_numeric_characteristic(self.IPA_extra_short, "+extra_short")

        # if we want more verbosity :
        if not IPA_only:
            res += str(self.abstract_description)

            # and we can ask the database for more informations :
            infos = IPA_data.database.informations_about_a_phoneme(self.IPA_articulation)
            s = "; ".join( "'{0}':'{1}'".format( key, infos[key] ) for key in infos)
            res += "DATABASE:{"+s+"}"

        return res

################################################################################
class PhoneticSegment(list):
    """
        class PhoneticSegment

        A PhoneticSegment is made of Syllable objects.

        The __init__ function HAS NO ARGUMENT.
    """

    IPA_secondary_articulations = []
    for l in (IPA_data.IPA_components[name] for name in IPA_data.IPA_components if name not in ("tone","stress")):
        for c in l:
            IPA_secondary_articulations.append(c)

    IPA_phonemes = "(?:"+\
                   "|"+"|".join( c for c in IPA_data.IPA_components['articulation'] )+"|"+\
                   "|"+"|".join( c for c in (l for l in IPA_secondary_articulations ))+"|"\
                   ")*"

    # (s,uː,) ˌ(v,ə,) '(n,i,ʁ)
    global_pattern = re.compile( "[{0}]?[{1}]?\({2},{2},{2}\)".format( "".join(IPA_data.IPA_components['tone']),
                                                                       "".join(IPA_data.IPA_components['stress']),
                                                                       IPA_phonemes) )
    # '(n,i,ʁ)
    syllabic_pattern = re.compile( "(?P<tone>[{0}]?)(?P<stress>[{1}]?)\((?P<onset>{2}),(?P<peak>{2}),(?P<coda>{2})\)".format( "".join(IPA_data.IPA_components['tone']),
                                                                                                                              "".join(IPA_data.IPA_components['stress']),
                                                                                                                              IPA_phonemes) )
    # "tabuːla"
    # (1) basic articulations : vowels (iyɨʉɯ...), consonants (ɱmnɳ...)
    #     without any diacritic
    phonetic_pattern__articulation      = "(?P<articulation><?:{0}>)".format( "|"+"|".join(IPA_data.IPA_components['articulation'])+"|")
    phonetic_pattern__tone              = "(?P<tone>[{0}]?)".format( "|"+"|".join(IPA_data.IPA_components['tone'])+"|")
    phonetic_pattern__stress            = "(?P<stress>[{0}]?)".format( "|"+"|".join(IPA_data.IPA_components['stress'])+"|")
    # (2) diacritics that go under the character :
    phonetic_pattern__advanced          = "(?P<advanced>[{0}]?)".format( "|"+"|".join(IPA_data.IPA_components['advanced'] )+"|")
    phonetic_pattern__advanced_tongue_root = "(?P<advanced_tongue_root>[{0}]?)".format( "|"+"|".join(IPA_data.IPA_components['advanced_tongue_root'] )+"|")
    phonetic_pattern__apical            = "(?P<apical>[{0}]?)".format( "|"+"|".join(IPA_data.IPA_components['apical'] )+"|")
    phonetic_pattern__breathy_voice     = "(?P<breathy_voice>[{0}]?)".format( "|"+"|".join(IPA_data.IPA_components['breathy_voice'] )+"|")
    phonetic_pattern__creaky_voice      = "(?P<creaky_voice>[{0}]?)".format( "|"+"|".join(IPA_data.IPA_components['creaky_voice'] )+"|")

    phonetic_pattern__dental            = "(?P<dental>[{0}]?)".format( "|"+"|".join(IPA_data.IPA_components['dental'] )+"|")
    phonetic_pattern__laminal           = "(?P<laminal>[{0}]?)".format( "|"+"|".join(IPA_data.IPA_components['laminal'] )+"|")
    phonetic_pattern__less_rounded      = "(?P<less_rounded>[{0}]?)".format( "|"+"|".join(IPA_data.IPA_components['less_rounded'] )+"|")
    phonetic_pattern__linguolabial_consonant = "(?P<linguolabial_consonant>[{0}]?)".format( "|"+"|".join(IPA_data.IPA_components['linguolabial_consonant'] )+"|")
    phonetic_pattern__lowered           = "(?P<lowered>[{0}]?)".format( "|"+"|".join(IPA_data.IPA_components['lowered'] )+"|")
    phonetic_pattern__mid_centralized   = "(?P<mid_centralized>[{0}]?)".format( "|"+"|".join(IPA_data.IPA_components['mid_centralized'] )+"|")
    phonetic_pattern__more_rounded      = "(?P<more_rounded>[{0}]?)".format( "|"+"|".join(IPA_data.IPA_components['more_rounded'] )+"|")
    phonetic_pattern__raised            = "(?P<raised>[{0}]?)".format( "|"+"|".join(IPA_data.IPA_components['raised'] )+"|")
    phonetic_pattern__retracted         = "(?P<retracted>[{0}]?)".format( "|"+"|".join(IPA_data.IPA_components['retracted'] )+"|")
    phonetic_pattern__retracted_tongue_root = "(?P<retracted_tongue_root>[{0}]?)".format( "|"+"|".join(IPA_data.IPA_components['retracted_tongue_root'] )+"|")
    phonetic_pattern__syllabic          = "(?P<syllabic>[{0}]?)".format( "|"+"|".join(IPA_data.IPA_components['syllabic'] )+"|")
    phonetic_pattern__voiced            = "(?P<voiced>[{0}]?)".format( "|"+"|".join(IPA_data.IPA_components['voiced'] )+"|")
    phonetic_pattern__voiceless         = "(?P<voiceless>[{0}]?)".format( "|"+"|".join(IPA_data.IPA_components['voiceless'] )+"|")
    # (3) diacritics that go over the character :
    phonetic_pattern__centralized_vowel = "(?P<centralized_vowel>[{0}]?)".format( "|"+"|".join(IPA_data.IPA_components['centralized_vowel'] )+"|")
    phonetic_pattern__nasalized         = "(?P<nasalized>[{0}]?)".format( "|"+"|".join(IPA_data.IPA_components['nasalized'] )+"|")
    phonetic_pattern__no_audible_release = "(?P<no_audible_release>[{0}]?)".format( "|"+"|".join(IPA_data.IPA_components['no_audible_release'] )+"|")
    # (4) diacritics that go beside the character :
    phonetic_pattern__aspirated         = "(?P<aspirated>[{0}]?)".format( "|"+"|".join(IPA_data.IPA_components['aspirated'])+"|")
    phonetic_pattern__labialized        = "(?P<labialized>[{0}]?)".format( "|"+"|".join(IPA_data.IPA_components['labialized'])+"|")
    phonetic_pattern__labio_palatalized = "(?P<labio_palatalized>[{0}]?)".format( "|"+"|".join(IPA_data.IPA_components['labio_palatalized'])+"|")
    phonetic_pattern__lateral_release   = "(?P<lateral_release>[{0}]?)".format( "|"+"|".join(IPA_data.IPA_components['lateral_release'])+"|")
    phonetic_pattern__length            = "(?P<length>[{0}]?)".format( "|"+"|".join( "|"+"|".join(IPA_data.IPA_components['length'])+"|"))
    phonetic_pattern__nasal_release     = "(?P<nasal_release>[{0}]?)".format( "|"+"|".join(IPA_data.IPA_components['nasal_release'])+"|")
    phonetic_pattern__palatalized       = "(?P<palatalized>[{0}]?)".format( "|"+"|".join(IPA_data.IPA_components['palatalized'])+"|")
    phonetic_pattern__pharyngealized    = "(?P<pharyngealized>[{0}]?)".format( "|"+"|".join(IPA_data.IPA_components['pharyngealized'])+"|")
    phonetic_pattern__rhotacized        = "(?P<rhotacized>[{0}]*)".format( "|"+"|".join(IPA_data.IPA_components['rhotacized'])+"|")
    phonetic_pattern__velarized         = "(?P<velarized>[{0}]?)".format( "|"+"|".join(IPA_data.IPA_components['velarized'])+"|")
    phonetic_pattern__voiced_aspirated  = "(?P<voiced_aspirated>[{0}]?)".format( "|"+"|".join(IPA_data.IPA_components['voiced_aspirated'])+"|")
    phonetic_pattern__glottalized       = "(?P<glottalized>[{0}]?)".format( "|"+"|".join(IPA_data.IPA_components['glottalized'])+"|")
    phonetic_pattern__extra_short       = "(?P<extra_short>[{0}]?)".format( "|"+"|".join(IPA_data.IPA_components['extra_short'])+"|")

    phonetic_pattern__non_syllabic      = "(?P<non_syllabic>[{0}]?)".format( "|"+"|".join(IPA_data.IPA_components['non_syllabic'] )+"|")
    phonetic_pattern__affricate         = "(?P<affricate>[{0}]?)".format( "|"+"|".join(IPA_data.IPA_components['affricate'])+"|")

    # be very careful : the order is important :
    # (1) basic articulations (2) diacritics under (3) diacritics over) (4) diacritics beside
    phonetic_pattern__str = "".join( (phonetic_pattern__articulation,
                                      phonetic_pattern__tone,
                                      phonetic_pattern__stress,
                                      phonetic_pattern__advanced,
                                      phonetic_pattern__advanced_tongue_root,
                                      phonetic_pattern__apical,
                                      phonetic_pattern__breathy_voice,
                                      phonetic_pattern__creaky_voice,
                                      phonetic_pattern__dental,
                                      phonetic_pattern__laminal,
                                      phonetic_pattern__less_rounded,
                                      phonetic_pattern__linguolabial_consonant,
                                      phonetic_pattern__lowered,
                                      phonetic_pattern__mid_centralized,
                                      phonetic_pattern__more_rounded,
                                      phonetic_pattern__raised,
                                      phonetic_pattern__retracted,
                                      phonetic_pattern__retracted_tongue_root,
                                      phonetic_pattern__syllabic,
                                      phonetic_pattern__voiced,
                                      phonetic_pattern__voiceless,
                                      phonetic_pattern__centralized_vowel,
                                      phonetic_pattern__nasalized,
                                      phonetic_pattern__no_audible_release,
                                      phonetic_pattern__aspirated,
                                      phonetic_pattern__labialized,
                                      phonetic_pattern__labio_palatalized,
                                      phonetic_pattern__lateral_release,
                                      phonetic_pattern__length,
                                      phonetic_pattern__nasal_release,
                                      phonetic_pattern__palatalized,
                                      phonetic_pattern__pharyngealized,
                                      phonetic_pattern__rhotacized,
                                      phonetic_pattern__velarized,
                                      phonetic_pattern__voiced_aspirated,
                                      phonetic_pattern__glottalized,
                                      phonetic_pattern__extra_short,
                                      phonetic_pattern__affricate,
                                      phonetic_pattern__non_syllabic,
                                      ))

    phonetic_pattern = re.compile( phonetic_pattern__str )

    #---------------------------------------------------------------------------
    def __repr__(self):
        """
                PhoneticSegment.__repr__
        """
        res = "PhoneticSegment : "

        if len(self)==0:
            return res+"(empty object)"
        else:
            return res+" |||| ".join( "syllable #{0}:{1}".format(i,repr(syllable)) for i,syllable in enumerate(self))

    #---------------------------------------------------------------------------
    def __str__(self):
        """
                PhoneticSegment.__str__
        """
        res = "PhoneticSegment : "

        if len(self)==0:
            return res+"(empty object)"
        else:
            return res+" |||| ".join( "syllable #{0}:{1}".format(i,str(syllable)) for i,syllable in enumerate(self))

    #---------------------------------------------------------------------------
    def add_a_SIPA_string_to_the_object(self,src):
        """
                PhoneticSegment.add_a_SIPA_string_to_the_object
        """

        #print("//add_a_SIPA_string_to_the_object//src//",src)

        # how much parentheses ?
        opened_parentheses = src.count( '(' )
        closed_parentheses = src.count( ')' )

        # ill-formed source string ?
        if opened_parentheses != closed_parentheses or\
           src.count(",") != 2:
            raise errors.PhoneticSegmentError_wrong_format_of_a_SIPA_string(SIPA_string=src)

        # main loop :
        syllable_counter = 0
        for syllable in re.finditer(PhoneticSegment.global_pattern, src):

            #print("//add_a_SIPA_string_to_the_object//syllable//",syllable)
            syllable_counter += 1

            for element_of_a_syllabe in re.finditer(PhoneticSegment.syllabic_pattern, syllable.group()):

                #print("__", element_of_a_syllabe.group())
                #print("coda",element_of_a_syllabe.group("coda"))

                tone = IPA_data.IPA_tone__character_to_value[element_of_a_syllabe.group("tone")]
                stress = IPA_data.IPA_stress__character_to_value[element_of_a_syllabe.group("stress")]

                onset = self.get_a_list_of_IPAphonemes_from_an_IPA_string( element_of_a_syllabe.group("onset") )
                peak = self.get_a_list_of_IPAphonemes_from_an_IPA_string( element_of_a_syllabe.group("peak") )
                coda = self.get_a_list_of_IPAphonemes_from_an_IPA_string( element_of_a_syllabe.group("coda") )

                new_syllable = Syllable(tone, stress, onset, peak, coda)
                self.append( new_syllable )

        if syllable_counter!=opened_parentheses:
            raise errors.PhoneticSegmentError_missing_syllable_in_a_SIPA_string(SIPA_string=src)

    #---------------------------------------------------------------------------
    def characters_substitution_in_an_input_string(self,src):
        """
                PhoneticSegment.characters_substitution_in_an_input_string

                PLEASE MODIFY THE DOCUMENTATION IF YOU MODIFY THIS FUNCTION :
                documentation/Sphinx/manually_generated_RST_files/IPA_characters
        """

        res = src

        #. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .
        # substitutions for exponants placed BEFORE an IPA character relative
        # to an articulation (L-exponants with L as left)
        #
        # ⁿk → n̆͡k
        #. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .
        list_of_Lexponants = { "ᵇ"      : "b",
                               "ʷ"      : "w",
                               "ⁿ"      : "n",
                               "ᵖ"      : "p",
                               "ᶢ"      : "g",
                               "ᵈ"      : "d"}

        # <exponants> : a dict of (int)index : (int)length of the exponant, (str)exponant
        p_articulations = "|".join( c for c in IPA_data.IPA_components['articulation'] )
        p_Lexponants = "|".join( c for c in list_of_Lexponants.keys())
        pattern__articulations_and_Lexponants__str = "([{0}]+({1}))".format(p_Lexponants,
                                                                            p_articulations)
        pattern__articulations_and_Lexponants = re.compile( pattern__articulations_and_Lexponants__str )
        pattern__Lexponants = re.compile( "{0}".format(p_Lexponants) )
        exponants = {}
        _res = [[c] for c in res]
        for r in re.finditer(pattern__articulations_and_Lexponants, res):

            r_group = r.group() # a string like "ᵇᵈk"
            index0 = r.start()

            for rr in re.finditer( pattern__Lexponants, r_group ):
                _res[index0+rr.start()] = [ list_of_Lexponants[rr.group()],
                                            '\u0306\u0361' ]

        res = "".join( "".join(l) for l in _res )

        #. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .
        # general substitutions :
        #. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .
        substitutions = {
                          # problem with the order of the diacritics :
                          "t̪ʼ"  : "tʼ̪", # t̪ʼ (0074+032A 02BC) → tʼ̪ (0074 02BC+032A)

                          # some character(s) that cannot be defined in the
                          # database since they use specific characteristics.
                          # E.g. "ɫ" stands for "lˠ" but the velarization
                          # cannot be described in the database.
                          "ɫ"   : "lˠ", # ɫ (026B) → lˠ (006C 02E0)

                          "ʨ"   : "t͡ɕ", # ʨ (02A8) → t͡ɕ (0074+0361 0255)

                          "ȷ̃"   : "j̃",  # ȷ̃ (0237 0303) → j̃ (006A 0303)

                          "ǃˢ"  : "ǃ͡s", # ǃˢ (01C3 02E2)→ ǃ͡s 01C3 0361 203C (http://en.wikipedia.org/wiki/Postalveolar_click)
                          "ᶢǃ͡s"  : "ǃ͡s", # ᶢǃ͡s (1DA2 01C3 0361 0073) → ǃ͡s 01C3 0361 203C (http://en.wikipedia.org/wiki/Postalveolar_click)
                          "ʗ̃ˢ"  : "ǃ͡s",  # ʗ̃ˢ (0297 0303 02E2) → ǃ͡s 01C3 0361 203C (http://en.wikipedia.org/wiki/Postalveolar_click)
                          "ᵑǃ͡sˀ"  : "ǃ͡s", # ᵑǃ͡sˀ (1D51 01C3+0361 0073 02C0) → ǃ͡s 01C3 0361 203C (http://en.wikipedia.org/wiki/Postalveolar_click)

                          # http://en.wikipedia.org/wiki/Dental_nasal_click :
                          "ᵑǀˀ"    : "ǀ̃ˀ", # ᵑǀˀ (1D51 01C0 02C0) → ǀ̃ˀ (01C0+0303 02C0)
                          "ᵑ̊ǀˀ"    : "ǀ̃ˀ", # ᵑ̊ǀˀ (1D51+030A 01C0 02C0) → ǀ̃ˀ (01C0+0303 02C0)
                          "ʇ̃ˀ"     : "ǀ̃ˀ", # ʇ̃ˀ (0287+0303 02C0) → ǀ̃ˀ (01C0+0303 02C0)
                          "ᵑʇˀ"    : "ǀ̃ˀ", # ᵑʇˀ (1D51 0287 02C0) → ǀ̃ˀ (01C0+0303 02C0)
                          "ᵑ̊ʇˀ"    : "ǀ̃ˀ", # ᵑ̊ʇˀ (1D51+030A 0287) → ǀ̃ˀ (01C0+0303 02C0)

                          # http://en.wikipedia.org/wiki/Alveolar_lateral_nasal_click
                          "ᵑǁ͡ʔ"  : "ǁ̃ˀ", # ᵑǁ͡ʔ (1D51 01C1+0361 0294) → ǁ̃ˀ (01C1 0303 02C0)
                          "ᵑ̊ǁˀ"  : "ǁ̃ˀ", # ᵑ̊ǁˀ (1D51+030A 02C0) → ǁ̃ˀ (01C1 0303 02C0)
                          "ʖ̃͜ʔ"  : "ǁ̃ˀ", # ʖ̃͜ʔ (0296+0303+035C 0294) → ǁ̃ˀ (01C1 0303 02C0)
                          "ᵑʖˀ"  : "ǁ̃ˀ", # ᵑʖˀ (1D51 0296 02C0) → ǁ̃ˀ (01C1 0303 02C0)

                          # http://en.wikipedia.org/wiki/Palatal_nasal_click
                          "ᵑǂ͡ʔ" : "ǂ̃ˀ", # ᵑǂ͡ʔ (1D51 01C2+0361 0294) →  ǂ̃ˀ (01C2+0303 02C0)
                          "ᵑ̊ǂˀ" : "ǂ̃ˀ", # ᵑ̊ǂˀ (1D5A+030A 01C2 02C0) →  ǂ̃ˀ (01C2+0303 02C0)
                          "ᵑǂˀ" : "ǂ̃ˀ", # ᵑǂˀ (1D51 01C2 02C0) →  ǂ̃ˀ (01C2+0303 02C0)

                          # http://en.wikipedia.org/wiki/Glottalized_bilabial_nasal_click#Glottalized_bilabial_nasal_click
                          "ʘ̃͜ʔ" : "ʘ̃ˀ", # ʘ̃͜ʔ (0298+0303+035C 0294) → ʘ̃ˀ (0298+0303 02C0)
                          "ᵑʘˀ" : "ʘ̃ˀ", # ᵑʘˀ (1D51 0298 02C0) → ʘ̃ˀ (0298+0303 02C0)
                          "ᵑ̊ʘˀ" : "ʘ̃ˀ", # ᵑ̊ʘˀ (1D51+030A 0298 02C0) → ʘ̃ˀ (0298+0303 02C0)

                          # http://en.wikipedia.org/wiki/Glottalized_alveolar_nasal_click#Glottalized_alveolar_nasal_click
                          "ᵑǃ͡ʔ" : "ǃ̃ˀ", # ᵑǃ͡ʔ (1D51 01C3+0361 0294) → ǃ̃ˀ (01C3 0303 02C0)
                          "ᵑ̊ǃˀ" : "ǃ̃ˀ", # ᵑ̊ǃˀ (1D51+030A 01C3 02C0) → ǃ̃ˀ (01C3 0303 02C0)
                          "ʗ̃͡ʔ"  : "ǃ̃ˀ", # ʗ̃͡ʔ (0297+0303+0361 0294) → ǃ̃ˀ (01C3 0303 02C0)
                          "ᵑʗˀ" : "ǃ̃ˀ", # ᵑʗˀ (1D51 0297 02C0) → ǃ̃ˀ (01C3 0303 02C0)

                          # diacritic for "ejective" phonemes :
                          "\u2019" : "\u02BC",

                          # varia :
                          "kʲʼ" : "kʼʲ",
                          "kʷʼ" : "kʼʷ",

                          "ɚ" : "ə˞",
                          "ɝ" : "ɜ˞",
                        }

        for key in substitutions:
            res = res.replace( key, substitutions[key] )

        return res

    #---------------------------------------------------------------------------
    def get_a_list_of_IPAphonemes_from_an_IPA_string(self,src):
        """
                PhoneticSegment.get_a_list_of_IPAphonemes_from_an_IPA_string

                "nid̥ʰa" → [ Phoneme('n'),Phoneme('i'),Phoneme('d̥ʰ'),Phoneme('a') ]

                This function returns a list of Phoneme objects made of one
                or more SimplePhoneme objects.
        """

        def delete_the_elements_of_a_list_marked_by_None(l):
            # This function modifies the list <l> IN PLACE !
            #
            # We erase elements in the list marked by None. But, since we read and
            # erase a list simultaneously we have to sort the indexes,
            # then to erase the last index, then the last index - 1, and so on.
            # The list shrinks by the end and the indexes point always to the right
            # place.
            indexes_to_be_erased = [index for index,phoneme in enumerate(l) if l[index] is None]
            indexes_to_be_erased.sort(reverse=True)
            for index in indexes_to_be_erased:
                l.pop(index)

        def first_nonNone_element(l,index,step):
            """
                The function searches an element in the list <l>
                before or after <index>; the searched element can't be None.
                The function returns None if no element matched
                or an integer if an element matches.
                The search begins at index + step, then index + step*2, ...

                step : +1 or -1; NO OTHER VALUES.
            """
            res = None

            if step==-1 and index<=0:
                return res

            if step==+1 and index>=len(l)-1:
                return res

            finished = False
            i = index+step

            while not finished:
                if l[i] is not None:
                    res = i
                    finished = True
                else:
                    # we have to go farther, at left :
                    if (step==-1 and i==0) or (step==+1 and i==len(l)-1):
                        # the loop is over.
                        finished = True
                    else:
                        i += step

            return res

        #. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .
        # We create a very basic list : each Phoneme object has ONE SimplePhoneme
        # object; for the moment we doesn't take in account diphtongs/triphtongs
        # or affricates.
        #. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .
        res = []
        for phoneme in re.finditer(PhoneticSegment.phonetic_pattern, src):

            # { "articulation:phoneme.group('articulation')
            #   ... }
            characteristics = dict( (name,phoneme.group(name)) for name in IPA_data.IPA_components )

            res.append( Phoneme( SimplePhoneme().init_from_the_IPA_characteristics(characteristics) ) )

        #. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .
        # Do we have affricates ?
        # If so, <affricates> stores the index of the first element.
        #. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .
        affricates = []
        for index,_phoneme in enumerate(res):
            phoneme = _phoneme[0]
            if phoneme.IPA_affricate==True:
                affricates.append(index)

        # we use the indexes <affricates> by begiginning with the highest index;
        # w use reversed-oder :
        affricates.reverse()
        for index in affricates:

            try:
                # we add to res[index] the next SimplePhoneme(s) stored in
                # the Phoneme object that is at position res[index+1] :
                res[index].append( res[index+1] )
                # we mark the (old) second SimplePhoneme in order to
                # erase it :
                res[index+1] = None
            except IndexError as e:
                raise errors.PhoneticSegmentError_affricate_with_only_one_element(src)

        delete_the_elements_of_a_list_marked_by_None(res)

        #. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .
        # Do we have non syllabic elements (=di-,tri-phthongs) ?
        # If so, <non_syllabic> stores the index of the first element.
        #. . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .
        non_syllabic = []
        for index,_phoneme in enumerate(res):
            phoneme = _phoneme[0]
            if phoneme.IPA_non_syllabic==True:
                non_syllabic.append(index)

        for i in non_syllabic:
            # we search where the Phoneme <i> (a non syllabic element)
            # has to be stored : before or after <i> ?

            # before <i> ?
            index_e = first_nonNone_element(res,i,-1)
            if index_e is not None and res[index_e].is_vocalic()==True:
                destination = index_e

            # after i ?
            index_e = first_nonNone_element(res,i,1)
            if index_e is not None and res[index_e].is_vocalic()==True:
                destination = index_e

            # neither ? something's wrong :
            if destination is None:
                raise errors.PhoneticSegmentError_pending_non_syllabic_element(src)

            # we add <i> to <destination> :
            # id est, we add to the Phoneme object res[destination]
            # the SimplePhoneme object res[i][x] :
            if i>destination:
                for simplephoneme in res[i]:
                    res[destination].append( simplephoneme )
            else:
                for simplephoneme in reversed(res[i]):
                    res[destination].insert( 0, simplephoneme )

            # we mark the <i> as to be erased :
            res[i] = None

        delete_the_elements_of_a_list_marked_by_None(res)

        return res

    #---------------------------------------------------------------------------
    def init_from_SIPA_strings(self,*src):
        """
                PhoneticSegment.init_from_SIPA_strings

                SIPA : Syllabic IPA

                format of a SIPA string : see the documentation
        """

        opened_parentheses = closed_parentheses = 0
        for _string in src:

            _string = self.characters_substitution_in_an_input_string(_string)

            try:

                # why should we use the (unicode) normalization 'NFD' ? We use this
                # function in order to get the 'canonical decomposition' of <string>.
                # With some diacritics (e.g. in ẽ, the nasalized one, code 0x303) the
                # regex won't work without this trick.
                string = unicodedata.normalize('NFD',_string)

                self.add_a_SIPA_string_to_the_object( string )

            except Exception as exc:

                # maybe an unknown character in <string>... ?
                for c in string:
                    if c not in IPA_data.IPA_components__all_characters+IPA_data.IPA_external_characters:
                        raise errors.PhoneticSegmentError_unknown_character_in_a_SIPA_string(SIPA_string = string,
                                                                                             unknown_character = c)

                # ... no, but something went wrong and we didn't know what was the problem...
                raise exc

        return self

    #---------------------------------------------------------------------------
    def phoneme_number(self,n):
        """
                PhoneticSegment.phoneme_number

                The function returnus the n-th phoneme in <self> with some
                informations about the syllable of the phoneme.

                res = ( <Phoneme> object,
                        index of syllable,
                        "onset|peak|coda",
                        index of the phoneme in the onset/peak/coda )
        """

        res = None

        i = -1
        found = False

        for i_syllable, syllable in enumerate(self):

            for i_phoneme,phoneme in enumerate(syllable.onset):
                i += 1

                if i==n:
                    res = (phoneme,i_syllable,'onset',i_phoneme)
                    found = True
                    break

            if found:
                break

            for i_phoneme,phoneme in enumerate(syllable.peak):
                i += 1

                if i==n:
                    res = (phoneme,i_syllable,'peak',i_phoneme)
                    found = True
                    break

            if found:
                break

            for i_phoneme,phoneme in enumerate(syllable.coda):
                i += 1

                if i==n:
                    res = (phoneme,i_syllable,'coda',i_phoneme)
                    found = True
                    break

            if found:
                break

        if res is None:
            raise IndexError("PhoneticSegment.phoneme_number : index {0} too high in {1}".format(n,self))
        return res
